# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Kernel manager that optionally seeds kernel memory."""

from tornado import gen
from notebook.services.kernels.kernelmanager import MappingKernelManager
from jupyter_client.ioloop import IOLoopKernelManager
from ..cell.parser import APICellParser

class SeedingMappingKernelManager(MappingKernelManager):
    """Extends the notebook kernel manager to optionally execute the contents
    of a notebook on a kernel when it starts.
    """
    def _kernel_manager_class_default(self):
        return 'kernel_gateway.services.kernels.manager.KernelGatewayIOLoopKernelManager'

    @property
    def seed_kernelspec(self):
        """Gets the kernel spec name required to run the seed notebook.

        Returns
        -------
        str
            Name of the notebook kernelspec or None if no seed notebook exists
        """
        if hasattr(self, '_seed_kernelspec'):
            return self._seed_kernelspec

        if self.parent.seed_notebook:
            self._seed_kernelspec = self.parent.seed_notebook['metadata']['kernelspec']['name']
        else:
            self._seed_kernelspec = None

        return self._seed_kernelspec

    @property
    def seed_source(self):
        """Gets the source of the seed notebook in cell order.

        Returns
        -------
        list
            Notebook code cell contents or None if no seed notebook exists
        """
        if hasattr(self, '_seed_source'):
            return self._seed_source

        if self.parent.seed_notebook:
            self._seed_source = [
                cell['source'] for cell in self.parent.seed_notebook.cells
                if cell['cell_type'] == 'code'
            ]
        else:
            self._seed_source = None

        return self._seed_source

    def start_seeded_kernel(self, *args, **kwargs):
        """Starts a kernel using the language specified in the seed notebook.

        If there is no seed notebook, start a kernel using the other parameters
        specified.
        """
        self.start_kernel(kernel_name=self.seed_kernelspec, *args, **kwargs)

    @gen.coroutine
    def start_kernel(self, *args, **kwargs):
        """Starts a kernel and then executes a list of code cells on it if a
        seed notebook exists.
        """
        kernel_id = yield gen.maybe_future(super(SeedingMappingKernelManager, self).start_kernel(*args, **kwargs))

        if kernel_id and self.seed_source is not None:
            # Only run source if the kernel spec matches the notebook kernel spec
            kernel = self.get_kernel(kernel_id)
            if kernel.kernel_name == self.seed_kernelspec:
                # Create a client to talk to the kernel
                client = kernel.client()
                # Only start channels and wait for ready in HTTP mode
                client.start_channels()
                client.wait_for_ready()
                for code in self.seed_source:
                    # Execute every non-API code cell and wait for each to
                    # succeed or fail
                    api_cell_parser = APICellParser(self.seed_kernelspec)
                    if not api_cell_parser.is_api_cell(code) and not api_cell_parser.is_api_response_cell(code):
                        client.execute(code)
                        msg = client.shell_channel.get_msg(block=True)
                        if msg['content']['status'] != 'ok':
                            # Shutdown the channels to remove any lingering ZMQ messages
                            client.stop_channels()
                            # Shutdown the kernel
                            self.shutdown_kernel(kernel_id)
                            raise RuntimeError('Error seeding kernel memory')
                # Shutdown the channels to remove any lingering ZMQ messages
                client.stop_channels()
        raise gen.Return(kernel_id)

class KernelGatewayIOLoopKernelManager(IOLoopKernelManager):
    """Extends the IOLoopKernelManager used by the SeedingMappingKernelManager
    to include the environment variable 'KERNEL_GATEWAY' set to '1', indicating
    that the notebook is executing within a Jupyter Kernel Gateway
    """
    def _launch_kernel(self, kernel_cmd, **kw):
        kw['env']['KERNEL_GATEWAY'] = '1'
        return super(KernelGatewayIOLoopKernelManager, self)._launch_kernel(kernel_cmd, **kw)



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Tornado handlers for kernel specs."""

from tornado import web
from ...mixins import TokenAuthorizationMixin, CORSMixin, JSONErrorsMixin
import os

class BaseSpecHandler(CORSMixin, web.StaticFileHandler):
    """Exposes the ability to return specifications from static files"""
    @staticmethod
    def get_resource_metadata():
        """Returns the (resource, mime-type) for the handlers spec.
        """
        pass

    def initialize(self):
        """Initializes the instance of this class to serve files.

        The handler is initialized to server files from the directory
        where this module is defined.
        """
        web.StaticFileHandler.initialize(self, path=os.path.dirname(__file__))

    def get(self):
        """Handler for a get on a specific handler
        """
        resource_name, content_type = self.get_resource_metadata()
        self.set_header('Content-Type', content_type)
        return web.StaticFileHandler.get(self, resource_name)

    def options(self, **kwargs):
        """Method for properly handling CORS pre-flight"""
        self.finish()

class SpecJsonHandler(BaseSpecHandler):
    """Exposes a JSON swagger specification"""
    @staticmethod
    def get_resource_metadata():
        return ('swagger.json','application/json')

class APIYamlHandler(BaseSpecHandler):
    """Exposes a YAML swagger specification"""
    @staticmethod
    def get_resource_metadata():
        return ('swagger.yaml', 'text/x-yaml')


default_handlers = [
    ('/api/{}'.format(SpecJsonHandler.get_resource_metadata()[0]), SpecJsonHandler),
    ('/api/{}'.format(APIYamlHandler.get_resource_metadata()[0]), APIYamlHandler)
]



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Manager that tracks kernel activity."""

from traitlets.config.configurable import LoggingConfigurable

# Constants for activity keys in the store
LAST_MESSAGE_TO_CLIENT = 'last_message_to_client'
LAST_MESSAGE_TO_KERNEL = 'last_message_to_kernel'
LAST_TIME_STATE_CHANGED = 'last_time_state_changed'
BUSY = 'busy'
CONNECTIONS = 'connections'
LAST_CLIENT_CONNECT = 'last_client_connect'
LAST_CLIENT_DISCONNECT = 'last_client_disconnect'

# Initial values for all kernel activity keys
DEFAULT_ACTIVITY_VALUES = {
    LAST_MESSAGE_TO_CLIENT: None,
    LAST_MESSAGE_TO_KERNEL: None,
    LAST_TIME_STATE_CHANGED: None,
    BUSY: False,
    CONNECTIONS: 0,
    LAST_CLIENT_CONNECT: None,
    LAST_CLIENT_DISCONNECT: None
}

class ActivityManager(LoggingConfigurable):
    """Represents a store of activity values for kernels. Intended to be used as a
    singleton.

    Parameters
    ----------
    kernel_manager : jupyter_client.multikernelmanager.MultiKernelManager
        Manager instance responsible for the lifecycle of all kernels tracked
        by this class

    Attributes
    ----------
    values : dict
        Kernel ID keys to dictionary of tracked activity values
    kernel_manager : jupyter_client.multikernelmanager.MultiKernelManager
        See parameters.
    """
    def __init__(self, kernel_manager, *args, **kwargs):
        super(ActivityManager, self).__init__(*args, **kwargs)
        self.values = {}
        self.kernel_manager = kernel_manager

    def get_map_for_kernel(self, kernel_id):
        """Gets activity values for a kernel.

        Parameters
        ----------
        kernel_id : str
            Unique identifier for the kernel

        Returns
        -------
        dict
            Activity values for the kernel or defaults if the `kernel_id` is
            not tracked
        """
        if kernel_id not in self.kernel_manager:
            return DEFAULT_ACTIVITY_VALUES.copy()

        if not kernel_id in self.values:
            self.values[kernel_id] = DEFAULT_ACTIVITY_VALUES.copy()

        return self.values[kernel_id]

    def publish(self, kernel_id, activity_type, value=None):
        """Sets the `value` stored for `activity_type` for `kernel_id`.

        Parameters
        ----------
        kernel_id : str
            Unique identifier for the kernel
        activity_type : str
            Activity key to set
        value : any
            Value to set for the activity
        """
        self.get_map_for_kernel(kernel_id)[activity_type] = value

    def increment_activity(self, kernel_id, activity_type):
        """Increments the `activity_type` value (an `int`) for `kernel_id`.

        Parameters
        ----------
        kernel_id : str
            Unique identifier for the kernel
        activity_type : str
            Activity key to set

        Raises
        ------
        TypeError
            If the stored value is not an int
        """
        self.get_map_for_kernel(kernel_id)[activity_type] += 1

    def decrement_activity(self, kernel_id, activity_type):
        """Decrements the `activity_type` value (an `int`) for `kernel_id`.

        Parameters
        ----------
        kernel_id : str
            Unique identifier for the kernel
        activity_type : str
            Activity key to set

        Raises
        ------
        TypeError
            If the stored value is not an int
        """
        self.get_map_for_kernel(kernel_id)[activity_type] -= 1

    def remove(self, kernel_id):
        """Removes all activity values for `kernel_id`.

        Adds the `kernel_id` to the set of IDs to ignore to prevent activities
        during shutdown from causing this manager to being tracking the
        kernel again.

        Parameters
        ----------
        kernel_id : str
            Unique identifier for the kernel
        """
        if kernel_id in self.values:
            del self.values[kernel_id]

    def get(self):
        """Gets all tracked activities for all kernels.

        Returns
        -------
        dict
            Kernel ID keys, activity dictionary values
        """
        return self.values



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Tornado handlers for kernel CRUD and communication."""

import json
import tornado
import notebook.services.kernels.handlers as notebook_handlers
from tornado import gen
from functools import partial
from datetime import datetime
from ...mixins import TokenAuthorizationMixin, CORSMixin, JSONErrorsMixin
from ...services.activity.manager import (LAST_MESSAGE_TO_CLIENT,
    LAST_MESSAGE_TO_KERNEL, LAST_TIME_STATE_CHANGED, BUSY, CONNECTIONS,
    LAST_CLIENT_CONNECT, LAST_CLIENT_DISCONNECT)

class MainKernelHandler(TokenAuthorizationMixin,
                        CORSMixin,
                        JSONErrorsMixin,
                        notebook_handlers.MainKernelHandler):
    """Extends the notebook main kernel handler with token auth, CORS, and
    JSON errors.
    """
    @gen.coroutine
    def post(self):
        """Overrides the super class method to honor the max number of allowed
        kernels configuration setting and to support custom kernel environment
        variables for every request.

        Delegates the request to the super class implementation if no limit is
        set or if the maximum is not reached. Otherwise, responds with an error.

        Raises
        ------
        tornado.web.HTTPError
            402 Payment Required (for lack of a better HTTP error code and
            mimicking what Google does) if the limit is reached
        """
        max_kernels = self.settings['kg_max_kernels']
        if max_kernels is not None:
            km = self.settings['kernel_manager']
            kernels = km.list_kernels()
            if len(kernels) >= max_kernels:
                raise tornado.web.HTTPError(402, 'Resource Limit')

        # Try to get env vars from the request body
        model = self.get_json_body()
        if model is not None and 'env' in model:
            if not isinstance(model['env'], dict):
                raise tornado.web.HTTPError(400)
            # Whitelist to KERNEL_* args only
            env = {key: value for key, value in model['env'].items() if key.startswith('KERNEL_')}
            # No way to override the call to start_kernel on the kernel manager
            # so do a temporary partial (ugh)
            orig_start = self.kernel_manager.start_kernel
            self.kernel_manager.start_kernel = partial(self.kernel_manager.start_kernel, env=env)
            try:
                yield super(MainKernelHandler, self).post()
            finally:
                self.kernel_manager.start_kernel = orig_start
        else:
            yield super(MainKernelHandler, self).post()

    def get(self):
        """Overrides the super class method to honor the kernel listing
        configuration setting.

        Allows the request to reach the super class if listing is enabled.

        Raises
        ------
        tornado.web.HTTPError
            403 Forbidden if kernel listing is disabled
        """
        if not self.settings.get('kg_list_kernels'):
            raise tornado.web.HTTPError(403, 'Forbidden')
        else:
            super(MainKernelHandler, self).get()

    def options(self, **kwargs):
        """Method for properly handling CORS pre-flight"""
        self.finish()

class KernelHandler(TokenAuthorizationMixin,
                    CORSMixin,
                    JSONErrorsMixin,
                    notebook_handlers.KernelHandler):
    """Extends the notebook kernel handler with token auth, CORS, JSON
    errors, and kernel activity tracking.
    """

    @property
    def activity(self):
        return self.settings['activity_manager']

    def delete(self, kernel_id):
        """Override the super class method to stop tracking activity for a
        kernel that is being deleted.

        Parameters
        ----------
        kernel_id
            ID of the kernel to stop tracking
        """
        if self.settings.get('kg_list_kernels'):
            self.activity.remove(kernel_id)
        super(KernelHandler, self).delete(kernel_id)

    def options(self, **kwargs):
        """Method for properly handling CORS pre-flight"""
        self.finish()

class ZMQChannelsHandler(TokenAuthorizationMixin,
                         notebook_handlers.ZMQChannelsHandler):
    """Extends the notebook websocket to zmq handler with token auth and
    kernel activity tracking.
    """
    @property
    def activity(self):
        return self.settings['activity_manager']

    def open(self, kernel_id):
        """Overrides the super class method to track connections to a kenrel.

        Parameters
        ----------
        kernel_id : str
            Opening a connection to this kernel
        """
        if self.settings.get('kg_list_kernels'):
            self.activity.publish(self.kernel_id, LAST_CLIENT_CONNECT, datetime.now().isoformat())
            self.activity.increment_activity(self.kernel_id, CONNECTIONS)
        super(ZMQChannelsHandler, self).open(kernel_id)

    def on_close(self):
        """Overrides the super class method to track disconnections from a
        kernel.
        """
        if self.settings.get('kg_list_kernels'):
            self.activity.publish(self.kernel_id, LAST_CLIENT_DISCONNECT, datetime.now().isoformat())
            self.activity.decrement_activity(self.kernel_id, CONNECTIONS)
        super(ZMQChannelsHandler, self).on_close()

    def _on_zmq_reply(self, stream, msg_list):
        """Overrides the super class method to track communication activity
        from a kernel as well as kernel idle/busy status.

        Parameters
        ----------
        kernel_id : str
            Kernel sending the message
        """
        if self.settings.get('kg_list_kernels'):
            msg_metadata = json.loads(msg_list[3].decode('UTF-8'))
            # If the message coming back is a status message, we need to inspect it
            if msg_metadata['msg_type'] == 'status':
                msg_content = json.loads(msg_list[6].decode('UTF-8'))
                # If the status is busy, set the busy to True
                if msg_content['execution_state'] == 'busy':
                    self.activity.publish(self.kernel_id, BUSY, True)
                # Else if the status is idle, set the busy to False
                elif msg_content['execution_state'] == 'idle':
                    self.activity.publish(self.kernel_id, BUSY, False)
                # Record the time the state was changed
                self.activity.publish(self.kernel_id, LAST_TIME_STATE_CHANGED, datetime.now().isoformat())

            self.activity.publish(self.kernel_id, LAST_MESSAGE_TO_CLIENT, datetime.now().isoformat())
        super(ZMQChannelsHandler, self)._on_zmq_reply(stream, msg_list)

    def on_message(self, msg):
        """Overrides the super class method to track communication activity
        to a kernel.

        Parameters
        ----------
        msg : str
            Message sent to a kernel
        """
        if self.settings.get('kg_list_kernels'):
            self.activity.publish(self.kernel_id, LAST_MESSAGE_TO_KERNEL, datetime.now().isoformat())
        super(ZMQChannelsHandler, self).on_message(msg)

default_handlers = []
for path, cls in notebook_handlers.default_handlers:
    if cls.__name__ in globals():
        # Use the same named class from here if it exists
        default_handlers.append((path, globals()[cls.__name__]))
    else:
        # Gen a new type with CORS and token auth
        bases = (TokenAuthorizationMixin, CORSMixin, cls)
        default_handlers.append((path, type(cls.__name__, bases, {})))



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Session manager that keeps all its metadata in memory."""

import uuid
from tornado import web, gen
from traitlets.config.configurable import LoggingConfigurable
from ipython_genutils.py3compat import unicode_type

class SessionManager(LoggingConfigurable):
    """Simple implementation of the SessionManager interface that allows clients
    to associate basic metadata with a kernel.

    Parameters
    ----------
    kernel_manager : SeedingMappingKernelManager
        Used to start a kernel when creating a session

    Attributes
    ----------
    kernel_manager : SeedingMappingKernelManager
        Used to start a kernel when creating a session
    _sessions : list
        Sessions
    _columns : list
        Session metadata key names
    """
    def __init__(self, kernel_manager, *args, **kwargs):
        super(SessionManager, self).__init__(*args, **kwargs)
        self.kernel_manager = kernel_manager
        self._sessions = []
        self._columns = ['session_id', 'path', 'kernel_id']

    def session_exists(self, path):
        """Checks to see if the session with the given path value exists.

        Parameters
        ----------
        path : str
            Session path value to search on

        Returns
        -------
        bool
        """
        return bool([item for item in self._sessions if item['path'] == path])

    def new_session_id(self):
        """Creates a uuid for a new session."""
        return unicode_type(uuid.uuid4())

    @gen.coroutine
    def create_session(self, path=None, kernel_name=None):
        """Creates a session and returns its model.

        Launches a kernel and stores the session metadata for later lookup.

        Parameters
        ----------
        path : str
            Path value to store in the session metadata
        kernel_name : str
            Kernel spec name

        Returns
        -------
        dict
            Session model
        """
        session_id = self.new_session_id()
        # allow nbm to specify kernels cwd
        kernel_id = yield gen.maybe_future(self.kernel_manager.start_kernel(path=path,
                                                                            kernel_name=kernel_name))
        raise gen.Return(self.save_session(session_id, path=path,
                                           kernel_id=kernel_id))

    def save_session(self, session_id, path=None, kernel_id=None):
        """Saves the metadata for the session with the given `session_id`.

        Given a `session_id` (and any other of the arguments), this method
        appends a dictionary to the in-memory list of sessions.

        Parameters
        ----------
        session_id : str
            UUID for the session; this method must be given a session_id
        path : str
            Path for the given notebook
        kernel_id : str
            ID for the kernel associated with this session

        Returns
        -------
        dict
            Session model with `session_id`, `path`, and `kernel_id` keys
        """
        self._sessions.append({'session_id': session_id,
                               'path':path,
                               'kernel_id': kernel_id})

        return self.get_session(session_id=session_id)

    def get_session_by_key(self, key, val):
        """Gets the first session with the given key/value pair.

        Parameters
        ----------
        key : hashable
            Session metadata key to match
        value : any
            Session metadata value to match

        Returns
        -------
        dict
            Matching session model or None if not found
        """
        s = [item for item in self._sessions if item[key] == val]
        return None if not s else s[0]

    def get_session(self, **kwargs):
        """Returns the model for a particular session.

        Takes a keyword argument and searches for the value in the in-memory
        session store. Returns the entire session model.

        Parameters
        ----------
        **kwargs : keyword argument
            One of the key/value pairs from `_columns`

        Raises
        ------
        TypeError
            If there are no kwargs or none of them match a key/column used in
            the metadata
        tornado.web.HTTPError
            404 Not Found if no session matches the provided metadata

        Returns
        -------
        model : dict
            All the information from the session described by the kwarg
        """
        if not kwargs:
            raise TypeError("Must specify a column to query")

        for param in kwargs.keys():
            if param not in self._columns:
                raise TypeError("No such column: %r", param)

        #multiple columns are never passed into kwargs so just using the
        #first and only one.
        column = list(kwargs.keys())[0]
        row = self.get_session_by_key(column, kwargs[column])

        if not row:
            raise web.HTTPError(404, u'Session not found: %s' % kwargs[column])

        return self.row_to_model(row)

    def update_session(self, session_id, **kwargs):
        """Updates the values in the session store.

        Update the values of the session model with the given `session_id`
        with the values from the keyword arguments.

        Parameters
        ----------
        session_id : str
            UUID that identifies a session in the sqlite3 database
        **kwargs : str
            Key/value pairs to store

        Raises
        ------
        KeyError
            If no session matches the given `session_id`
        """
        if not kwargs:
            # no changes
            return

        row = self.get_session_by_key('session_id', session_id)

        if not row:
            raise KeyError

        self._sessions.remove(row)

        if 'path' in kwargs:
            row['path'] = kwargs['path']

        if 'kernel_id' in kwargs:
            row['kernel_id'] = kwargs['kernel_id']

        self._sessions.append(row)

    def row_to_model(self, row):
        """Turns a "row" in the in-memory session store into a model dictionary.

        Parameters
        ----------
        row : dict
            Maps `id` to `session_id`, `notebook` to a dict containing the
            `path`, and `kernel` to the kernel model looked up using the
            `kernel_id`
        """
        if row['kernel_id'] not in self.kernel_manager:
            # The kernel was killed or died without deleting the session.
            # We can't use delete_session here because that tries to find
            # and shut down the kernel.
            self._sessions.remove(row)
            raise KeyError

        model = {
            'id': row['session_id'],
            'notebook': {
                'path': row['path']
            },
            'kernel': self.kernel_manager.kernel_model(row['kernel_id'])
        }
        return model

    def list_sessions(self):
        """Returns a list of dictionaries containing all the information from
        the session store.

        Returns
        -------
        list
            Dictionaries from `row_to_model`
        """
        l = [self.row_to_model(r) for r in self._sessions]
        return l

    def delete_session(self, session_id):
        """Deletes the session in the session store with given `session_id`.

        Raises
        ------
        KeyError
            If the `session_id` is not in the store
        """
        # Check that session exists before deleting
        s = self.get_session_by_key('session_id', session_id)
        if not s:
            raise KeyError

        self.kernel_manager.shutdown_kernel(s['kernel_id'])
        self._sessions.remove(s)



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Kernel Gateway Jupyter application."""

import os
import logging
import nbformat

try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse

from traitlets import Unicode, Integer, Bool, default

from jupyter_core.application import JupyterApp
from jupyter_client.kernelspec import KernelSpecManager

# Install the pyzmq ioloop. This has to be done before anything else from
# tornado is imported.
from zmq.eventloop import ioloop
ioloop.install()

from tornado import httpserver
from tornado import web
from tornado.log import enable_pretty_logging

from .services.api.handlers import default_handlers as default_api_handlers
from .services.kernels.handlers import default_handlers as default_kernel_handlers
from .services.kernelspecs.handlers import default_handlers as default_kernelspec_handlers
from .services.sessions.handlers import default_handlers as default_session_handlers
from .services.sessions.sessionmanager import SessionManager
from .services.activity.manager import ActivityManager
from .services.kernels.manager import SeedingMappingKernelManager
from .services.kernels.pool import KernelPool, ManagedKernelPool
from .base.handlers import default_handlers as default_base_handlers
from .services.notebooks.handlers import NotebookAPIHandler, parameterize_path, NotebookDownloadHandler
from .services.cell.parser import APICellParser
from .services.swagger.handlers import SwaggerSpecHandler
from .services.activity.handlers import ActivityHandler

from notebook.utils import url_path_join

class KernelGatewayApp(JupyterApp):
    """Application that provisions Jupyter kernels and proxies HTTP/Websocket
    traffic to the kernels.

    - reads command line and environment variable settings
    - initializes managers and routes
    - creates a Tornado HTTP server
    - starts the Tornado event loop
    """
    name = 'jupyter-kernel-gateway'
    description = """
        Jupyter Kernel Gateway

        Provisions Jupyter kernels and proxies HTTP/Websocket traffic
        to them.
    """
    # Server IP / PORT binding
    port_env = 'KG_PORT'
    port = Integer(config=True,
        help="Port on which to listen (KG_PORT env var)"
    )
    @default('port')
    def port_default(self):
        return int(os.getenv(self.port_env, 8888))

    ip_env = 'KG_IP'
    ip = Unicode(config=True,
        help="IP address on which to listen (KG_IP env var)"
    )
    @default('ip')
    def ip_default(self):
        return os.getenv(self.ip_env, '127.0.0.1')

    # Base URL
    base_url_env = 'KG_BASE_URL'
    base_url = Unicode(config=True,
        help="""The base path for mounting all API resources (KG_BASE_URL env var)""")
    @default('base_url')
    def base_url_default(self):
        return os.getenv(self.base_url_env, '/')

    # Token authorization
    auth_token_env = 'KG_AUTH_TOKEN'
    auth_token = Unicode(config=True,
        help='Authorization token required for all requests (KG_AUTH_TOKEN env var)'
    )
    @default('auth_token')
    def _auth_token_default(self):
        return os.getenv(self.auth_token_env, '')

    # CORS headers
    allow_credentials_env = 'KG_ALLOW_CREDENTIALS'
    allow_credentials = Unicode(config=True,
        help='Sets the Access-Control-Allow-Credentials header. (KG_ALLOW_CREDENTIALS env var)'
    )
    @default('allow_credentials')
    def allow_credentials_default(self):
        return os.getenv(self.allow_credentials_env, '')

    allow_headers_env = 'KG_ALLOW_HEADERS'
    allow_headers = Unicode(config=True,
        help='Sets the Access-Control-Allow-Headers header. (KG_ALLOW_HEADERS env var)'
    )
    @default('allow_headers')
    def allow_headers_default(self):
        return os.getenv(self.allow_headers_env, '')

    allow_methods_env = 'KG_ALLOW_METHODS'
    allow_methods = Unicode(config=True,
        help='Sets the Access-Control-Allow-Methods header. (KG_ALLOW_METHODS env var)'
    )
    @default('allow_methods')
    def allow_methods_default(self):
        return os.getenv(self.allow_methods_env, '')

    allow_origin_env = 'KG_ALLOW_ORIGIN'
    allow_origin = Unicode(config=True,
        help='Sets the Access-Control-Allow-Origin header. (KG_ALLOW_ORIGIN env var)'
    )
    @default('allow_origin')
    def allow_origin_default(self):
        return os.getenv(self.allow_origin_env, '')

    expose_headers_env = 'KG_EXPOSE_HEADERS'
    expose_headers = Unicode(config=True,
        help='Sets the Access-Control-Expose-Headers header. (KG_EXPOSE_HEADERS env var)'
    )
    @default('expose_headers')
    def expose_headers_default(self):
        return os.getenv(self.expose_headers_env, '')

    max_age_env = 'KG_MAX_AGE'
    max_age = Unicode(config=True,
        help='Sets the Access-Control-Max-Age header. (KG_MAX_AGE env var)'
    )
    @default('max_age')
    def max_age_default(self):
        return os.getenv(self.max_age_env, '')

    max_kernels_env = 'KG_MAX_KERNELS'
    max_kernels = Integer(config=True,
        allow_none=True,
        help='Limits the number of kernel instances allowed to run by this gateway. (KG_MAX_KERNELS env var)'
    )
    @default('max_kernels')
    def max_kernels_default(self):
        val = os.getenv(self.max_kernels_env)
        return val if val is None else int(val)

    seed_uri_env = 'KG_SEED_URI'
    seed_uri = Unicode(config=True,
        allow_none=True,
        help='Runs the notebook (.ipynb) at the given URI on every kernel launched. (KG_SEED_URI env var)'
    )
    @default('seed_uri')
    def seed_uri_default(self):
        return os.getenv(self.seed_uri_env)

    prespawn_count_env = 'KG_PRESPAWN_COUNT'
    prespawn_count = Integer(config=True,
        default_value=None,
        allow_none=True,
        help='Number of kernels to prespawn using the default language. (KG_PRESPAWN_COUNT env var)'
    )
    @default('prespawn_count')
    def prespawn_count_default(self):
        val = os.getenv(self.prespawn_count_env)
        return val if val is None else int(val)

    default_kernel_name_env = 'KG_DEFAULT_KERNEL_NAME'
    default_kernel_name = Unicode(config=True,
        help="""The default kernel name when spawning a kernel (KG_DEFAULT_KERNEL_NAME env var)""")
    @default('default_kernel_name')
    def default_kernel_name_default(self):
        # defaults to Jupyter's default kernel name on empty string
        return os.getenv(self.default_kernel_name_env, '')

    list_kernels_env = 'KG_LIST_KERNELS'
    list_kernels = Bool(config=True,
        help="""Permits listing of the running kernels using API endpoints /api/kernels
            and /api/sessions (KG_LIST_KERNELS env var). Note: Jupyter Notebook
            allows this by default but kernel gateway does not."""
    )
    @default('list_kernels')
    def list_kernels_default(self):
        return os.getenv(self.list_kernels_env, 'False') == 'True'

    api_env = 'KG_API'
    api = Unicode('jupyter-websocket',
        config=True,
        help='Controls which API to expose, that of a Jupyter kernel or the seed notebook\'s, using values "jupyter-websocket" or "notebook-http" (KG_API env var)'
    )
    @default('api')
    def api_default(self):
        return os.getenv(self.api_env, 'jupyter-websocket')

    def _api_changed(self, name, old, new):
        if new not in ['notebook-http', 'jupyter-websocket']:
            raise ValueError('Invalid API value, valid values are jupyter-websocket and notebook-http')

    allow_notebook_download_env = 'KG_ALLOW_NOTEBOOK_DOWNLOAD'
    allow_notebook_download = Bool(
        config=True,
        help="Optional API to download the notebook source code in notebook-http mode, defaults to not allow"
    )
    @default('allow_notebook_download')
    def allow_notebook_download_default(self):
        return os.getenv(self.allow_notebook_download_env, 'False') == 'True'

    def _load_notebook(self, uri):
        """Loads a notebook from the local filesystem or HTTP URL.

        Raises
        ------
        RuntimeError if no installed kernel can handle the language specified
        in the notebook.

        Returns
        -------
        object
            Notebook object from nbformat
        """
        parts = urlparse(uri)

        if parts.netloc == '' or parts.netloc == 'file':
            # Local file
            with open(parts.path) as nb_fh:
                notebook = nbformat.read(nb_fh, 4)
        else:
            # Remote file
            import requests
            resp = requests.get(uri)
            resp.raise_for_status()
            notebook = nbformat.reads(resp.text, 4)

        # Error if no kernel spec can handle the language requested
        kernel_name = notebook['metadata']['kernelspec']['name']
        self.kernel_spec_manager.get_kernel_spec(kernel_name)

        return notebook

    def initialize(self, argv=None):
        """Initializes the base class, configurable manager instances, the
        Tornado web app, and the tornado HTTP server.

        Parameters
        ----------
        argv
            Command line arguments
        """
        super(KernelGatewayApp, self).initialize(argv)
        self.init_configurables()
        self.init_webapp()
        self.init_http_server()

    def init_configurables(self):
        """Initializes all configurable objects including a kernel manager, kernel
        spec manager, session manager, kernel pool, and activity manager.

        Optionally, loads a notebook and prespawns the configured number of
        kernels.
        """
        self.kernel_spec_manager = KernelSpecManager(parent=self)

        self.seed_notebook = None
        if self.seed_uri is not None:
            # Note: must be set before instantiating a SeedingMappingKernelManager
            self.seed_notebook = self._load_notebook(self.seed_uri)

        # Only pass a default kernel name when one is provided. Otherwise,
        # adopt whatever default the kernel manager wants to use.
        kwargs = {}
        if self.default_kernel_name:
            kwargs['default_kernel_name'] = self.default_kernel_name
        self.kernel_manager = SeedingMappingKernelManager(
            parent=self,
            log=self.log,
            connection_dir=self.runtime_dir,
            kernel_spec_manager=self.kernel_spec_manager,
            **kwargs
        )

        self.activity_manager = ActivityManager(
            parent=self,
            log=self.log,
            kernel_manager=self.kernel_manager
        )

        self.session_manager = SessionManager(
            log=self.log,
            kernel_manager=self.kernel_manager
        )
        self.contents_manager = None

        if self.prespawn_count:
            if self.max_kernels and self.prespawn_count > self.max_kernels:
                raise RuntimeError('cannot prespawn {}; more than max kernels {}'.format(
                    self.prespawn_count, self.max_kernels)
                )

        if self.api == 'notebook-http':
            self.kernel_pool = ManagedKernelPool(
                self.prespawn_count,
                self.kernel_manager
            )
        else:
            self.kernel_pool = KernelPool(
                self.prespawn_count,
                self.kernel_manager
            )

    def init_webapp(self):
        """Initializes Tornado web application with kernel handlers.

        Adds the various managers and web-front configuration values to the
        Tornado settings for reference by the handlers.

        Notes
        -----
        Uses the `api` setting to determine which handlers to add.
        Developers should note: this may be refactored in the future.
        """
        # Redefine handlers off the base_url path
        handlers = []
        if self.api == 'notebook-http':
            # Register the NotebookDownloadHandler if configuration allows
            if self.allow_notebook_download:
                handlers.append((
                    url_path_join('/', self.base_url, r'/_api/source'),
                    NotebookDownloadHandler,
                    {'path': self.seed_uri}
                ))

            # Discover the notebook endpoints and their implementations
            parser = APICellParser(self.kernel_manager.seed_kernelspec)
            endpoints = parser.endpoints(self.kernel_manager.seed_source)
            response_sources = parser.endpoint_responses(self.kernel_manager.seed_source)
            if len(endpoints) == 0:
                raise RuntimeError('No endpoints were discovered. Check your notebook to make sure your cells are annotated correctly.')

            # Cycle through the (endpoint_path, source) tuples and register their handlers
            for endpoint_path, verb_source_map in endpoints:
                parameterized_path = parameterize_path(endpoint_path)
                parameterized_path = url_path_join('/', self.base_url, parameterized_path)
                self.log.info('Registering endpoint_path: {}, methods: ({})'.format(
                    parameterized_path,
                    list(verb_source_map.keys())
                ))
                response_source_map = response_sources[endpoint_path] if endpoint_path in response_sources else {}
                handler_args = { 'sources' : verb_source_map,
                    'response_sources' : response_source_map,
                    'kernel_pool' : self.kernel_pool,
                    'kernel_name' : self.kernel_manager.seed_kernelspec
                }
                handlers.append((parameterized_path, NotebookAPIHandler, handler_args))

            # Register the swagger API spec handler
            handlers.append(
                (url_path_join('/', self.base_url, r'/_api/spec/swagger.json'),
                SwaggerSpecHandler, {
                    'notebook_path' : self.seed_uri,
                    'source_cells': self.kernel_manager.seed_source,
                    'kernel_spec' : self.kernel_manager.seed_kernelspec
            }))

            # Register the 404 catch-all last
            handlers.append(default_base_handlers[-1])

            # Enable the same pretty logging the notebook uses
            enable_pretty_logging()
        elif self.api == 'jupyter-websocket':
            # append the activity monitor for websocket mode
            handlers.append((
                url_path_join('/', self.base_url, r'/_api/activity'),
                ActivityHandler,
                {}
            ))
            # append tuples for the standard kernel gateway endpoints
            for handler in (
                default_api_handlers +
                default_kernel_handlers +
                default_kernelspec_handlers +
                default_session_handlers +
                default_base_handlers
            ):
                # Create a new handler pattern rooted at the base_url
                pattern = url_path_join('/', self.base_url, handler[0])
                # Some handlers take args, so retain those in addition to the
                # handler class ref
                new_handler = tuple([pattern] + list(handler[1:]))
                handlers.append(new_handler)

        # Configure the tornado logging level too
        logging.getLogger().setLevel(self.log_level)

        self.web_app = web.Application(
            handlers=handlers,
            activity_manager=self.activity_manager,
            kernel_manager=self.kernel_manager,
            session_manager=self.session_manager,
            contents_manager=self.contents_manager,
            kernel_spec_manager=self.kernel_manager.kernel_spec_manager,
            kg_auth_token=self.auth_token,
            kg_allow_credentials=self.allow_credentials,
            kg_allow_headers=self.allow_headers,
            kg_allow_methods=self.allow_methods,
            kg_allow_origin=self.allow_origin,
            kg_expose_headers=self.expose_headers,
            kg_max_age=self.max_age,
            kg_max_kernels=self.max_kernels,
            kg_list_kernels=self.list_kernels,
            kg_api=self.api,
            # Also set the allow_origin setting used by notebook so that the
            # check_origin method used everywhere respects the value
            allow_origin=self.allow_origin
        )

    def init_http_server(self):
        """Initializes a HTTP server for the Tornado web application on the
        configured interface and port.
        """
        self.http_server = httpserver.HTTPServer(self.web_app)
        self.http_server.listen(self.port, self.ip)

    def start(self):
        """Starts an IO loop for the application."""
        super(KernelGatewayApp, self).start()
        self.log.info('The Jupyter Kernel Gateway is running at: http://{}:{}'.format(
            self.ip, self.port
        ))

        self.io_loop = ioloop.IOLoop.current()

        try:
            self.io_loop.start()
        except KeyboardInterrupt:
            self.log.info("Interrupted...")

    def stop(self):
        """
        Stops the HTTP server and IO loop associated with the application.
        """
        def _stop():
            self.http_server.stop()
            self.io_loop.stop()
        self.io_loop.add_callback(_stop)

    def shutdown(self):
        """Stops all kernels in the pool."""
        self.kernel_pool.shutdown()

launch_instance = KernelGatewayApp.launch_instance



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Utilities for handling HTTP requests."""

import json
import re
from tornado.httputil import parse_body_arguments

_named_param_regex = re.compile('(:([^/\s]*))')
FORM_URLENCODED = 'application/x-www-form-urlencoded'
MULTIPART_FORM_DATA = 'multipart/form-data'
APPLICATION_JSON = 'application/json'
TEXT_PLAIN = 'text/plain'

def format_request(bundle):
    """Creates an assignment statement of bundle JSON-encoded to a variable
    named `REQUEST`.

    Returns
    -------
    str
        `REQUEST = "<json-encoded expression>"`
    """
    bundle = json.dumps(bundle)
    statement = "REQUEST = {}".format(bundle)
    return statement

def parameterize_path(path):
    """Creates a regex to match all named parameters in a path.

    Parameters
    ----------
    path : str
        URL path with `/:name` segements

    Returns
    -------
    str
        Path with `:name` parameters replaced with regex patterns for matching
        them.
    """
    matches = re.findall(_named_param_regex, path)
    for match in matches:
        path = path.replace(match[0], '(?P<{}>[^\/]+)'.format(match[1]))
    return path.strip()

def parse_body(request):
    """Parses the body of an HTTP request based on its Content-Type.

    If no Content-Type is found, treats the value as plain text.

    Parameters
    ----------
    request : tornado.web.HTTPRequest
        Web request with the body to parse

    Returns
    -------
    dict or str
        Dictionary of a form-encoded JSON-encoded body, raw string otherwise
    """
    content_type = TEXT_PLAIN
    body = request.body
    body = '' if body is b'' or body is None else body.decode(encoding='UTF-8')
    if 'Content-Type' in request.headers:
        content_type = request.headers['Content-Type']
    return_body = body
    if content_type == FORM_URLENCODED or content_type.startswith(MULTIPART_FORM_DATA):
        # If there is form data, we already have the values in body_arguments, we
        # just need to convert the byte arrays to strings
        return_body = parse_args(request.body_arguments)
    elif content_type == APPLICATION_JSON:
        # Trying parsing the json, if we can't parse it the initial assignment
        # will treat the body as text
        try:
            return_body = json.loads(body)
        except Exception:
            pass
    return return_body

def parse_args(args):
    """Decodes UTF-8 encoded argument values.

    Parameters
    ----------
    args : dict
        Maps arbitrary keys to UTF-8 encoded strings / byte-arrays

    Returns
    -------
    dict
        Maps keys from args to decoded strings
    """
    rv = {}
    for key in args:
        rv[key] = []
        for value in args[key]:
            rv[key].append(value.decode(encoding='UTF-8'))
    return rv

def headers_to_dict(headers):
    """Turns a set of tornado headers into a Python dict.

    Repeat headers are aggregated into lists.

    Parameters
    ----------
    headers : dict
        Key / value header pairs

    Returns
    -------
    dict
        Maps keys from headers to values or lists of values
    """
    new_headers = {}
    for header, header_value in headers.get_all():
        if header in new_headers:
            if not isinstance(new_headers[header], list):
                original_value = new_headers[header]
                new_headers[header] = [original_value]
            new_headers[header].append(header_value)
        else:
            new_headers[header] = header_value

    return new_headers



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Exception classes for notebook-http mode."""

class CodeExecutionError(Exception):
    """Raised when a notebook's code fails to execute in response to an API
    request.
    """
    pass

class UnsupportedMethodError(Exception):
    """Raised when a notebook-defined API does not support the requested HTTP
    method.
    """
    pass



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import os
import sys
from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))

version_ns = {}
with open(os.path.join(here, 'kernel_gateway', '_version.py')) as f:
    exec(f.read(), {}, version_ns)

setup_args = dict(
    name='jupyter_kernel_gateway',
    author='Jupyter Development Team',
    author_email='jupyter@googlegroups.com',
    url='http://github.com/jupyter-incubator/kernel_gateway',
    description='A web server for spawning and communicating with Jupyter kernels',
    long_description='''\
Jupyter Kernel Gateway is a web server that supports different mechanisms for
spawning and communicating with Jupyter kernels, such as:

* A Jupyter Notebook server-compatible HTTP API used for requesting kernels
  and talking the `Jupyter kernel protocol <https://jupyter-client.readthedocs.org/en/latest/messaging.html>`_
  with the kernels over Websockets
* A HTTP API defined by annotated notebook cells that maps HTTP verbs and
  resources to code to execute on a kernel

The server launches kernels in its local process/filesystem space. It can be
containerized and scaled out using common technologies like
`tmpnb <https://github.com/jupyter/tmpnb>`_,
`Cloud Foundry <https://github.com/cloudfoundry>`_, and
`Kubernetes <http://kubernetes.io/>`_.
''',
    version=version_ns['__version__'],
    license='BSD',
    platforms="Linux, Mac OS X, Windows",
    keywords=['Interactive', 'Interpreter', 'Kernel', 'Web', 'Cloud'],
    packages=[
        'kernel_gateway',
        'kernel_gateway.base',
        'kernel_gateway.services',
        'kernel_gateway.services.activity',
        'kernel_gateway.services.api',
        'kernel_gateway.services.cell',
        'kernel_gateway.services.kernels',
        'kernel_gateway.services.kernelspecs',
        'kernel_gateway.services.notebooks',
        'kernel_gateway.services.sessions',
        'kernel_gateway.services.swagger',
    ],
    scripts=[
        'scripts/jupyter-kernelgateway'
    ],
    install_requires=[
        'jupyter_core>=4.0,<5.0',
        'jupyter_client>=4.2.0,<5.0',
        'notebook>=4.1.0,<5.0',
        'traitlets>=4.2.0,<5.0',
        'requests>=2.7,<3.0'
    ],
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3'
    ],
)

if 'setuptools' in sys.modules:
    # setupstools turns entrypoint scripts into executables on windows
    setup_args['entry_points'] = {
        'console_scripts': [
            'jupyter-kernelgateway = kernel_gateway:launch_instance'
        ]
    }
    # Don't bother installing the .py scripts if if we're using entrypoints
    setup_args.pop('scripts', None)

if __name__ == '__main__':
    setup(**setup_args)



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
"""Tornado handler for kernel activity tracking."""

import tornado.web
import json
from ...mixins import TokenAuthorizationMixin, CORSMixin, JSONErrorsMixin

class ActivityHandler(TokenAuthorizationMixin,
                      CORSMixin,
                      JSONErrorsMixin,
                      tornado.web.RequestHandler):

    @property
    def activity(self):
        return self.settings['activity_manager']

    def get(self):
        """Responds with a JSON object of stats about every running kernel if
        kernel listing, `kg_list_kernels`, is enabled on the server.

        Raises
        ------
        tornado.web.HTTPError
            If kg_list_kernels is False, respond with 403 Forbidden

        See Also
        --------
        manager.ActivityManager : Tracks kernel activity

        Examples
        --------

        {
          "0f41b09c-c5b4-4f28-a7db-2f779151c20f": {
            "last_message_to_client": "2014-09-08T19:40:17.819Z",
            "last_message_to_kernel": "2014-09-08T19:40:17.819Z",
            "last_time_state_changed": "2014-09-08T19:40:17.819Z",
            "busy" : false,
            "connections" : 0,
            "last_client_connect": "2014-09-08T19:40:17.819Z",
            "last_client_disconnect": "2014-09-08T19:40:17.819Z"
          }
        }
        """
        if not self.settings.get('kg_list_kernels'):
            raise tornado.web.HTTPError(403, 'Forbidden')
        self.set_header('Content-Type', 'application/json')
        self.set_status(200)
        self.finish(json.dumps(self.activity.get()))
